!! ##############################################################################################
!!
!! Copyright 2012 CNRS, INPT
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!> @file qrm_analysis_mod.F90
!! This file contains a subroutine for computing the min-norm solution of a problem
!!
!! $Date: 2016-01-29 22:22:30 +0100 (Fri, 29 Jan 2016) $
!! $Author: abuttari $
!! $Version: 1.1$
!! $Revision: 2075 $
!!
!! ##############################################################################################

#include "qrm_common.h"

!> @brief This routine computes the min-norm solution of a problem

!> This routine computes the min-norm solution of an underdetermined system
!! Ax=b with multiple RHSs.
!!
!! @param[in]     qrm_mat a qrm_spmat_type data which contains the input matrix.
!!                        On output the original data will be unchanged and
!!                        the result of the analysis and factorization phases will
!!                        be stored in the adata and fdata fields, respectively.
!! @param[in]           b the RHSs. A 2D array of leading dimension qrm_mat%m. On output
!!                        it will contain b
!! @param[out]          x the solution, i.e., Q*R'\b
!! 
subroutine dqrm_min_norm2d(qrm_mat, b, x)
  use dqrm_spmat_mod
  use qrm_mem_mod
  use qrm_error_mod
  use dqrm_analysis_mod
  use dqrm_factorization_mod
  use dqrm_solve_mod
  implicit none

  type(dqrm_spmat_type) :: qrm_mat
  real(kind(1.d0)) :: b(:,:), x(:,:)

  ! error management
  integer                         :: err_act
  character(len=*), parameter     :: name='qrm_min_norm'

  call qrm_err_act_save(err_act)

  __QRM_PRNT_DBG('("Entering the min-norm driver")')

  call dqrm_check_spmat(qrm_mat)
  __QRM_CHECK_RET(name,'qrm_check_spmat',9999)
  
  if(qrm_mat%n .lt. qrm_mat%m) then
     call qrm_err_push(31, name,ied=(/qrm_mat%m,qrm_mat%n,0,0,0/))
     goto 9999
  end if

  ! analysis
  call dqrm_analyse(qrm_mat, 't')
  __QRM_CHECK_RET(name,'qrm_analyse',9999)
  ! factorization
  call dqrm_factorize(qrm_mat, 't')
  __QRM_CHECK_RET(name,'qrm_factorize',9999)

  call dqrm_solve2d(qrm_mat, 't', b, x)
  __QRM_CHECK_RET(name,'qrm_solve',9999)
  call dqrm_apply2d(qrm_mat, 'n', x)
  __QRM_CHECK_RET(name,'qrm_apply',9999)

  call qrm_err_act_restore(err_act)
  return
  
9999 continue ! error management
  call qrm_err_act_restore(err_act)
  if(err_act .eq. qrm_abort_) then
     call qrm_err_check()
  end if
  return
  
end subroutine dqrm_min_norm2d




!> @brief This routine computes the min-norm solution of a problem

!> This routine computes the min-norm solution of an underdetermined system
!! Ax=b with a single RHS.
!!
!! @param[in]     qrm_mat a qrm_spmat_type data which contains the input matrix.
!!                        On output the original data will be unchanged and
!!                        the result of the analysis and factorization phases will
!!                        be stored in the adata and fdata fields, respectively.
!! @param[in]           b the RHSs. A 1D array of leading dimension qrm_mat%m. On output
!!                        it will contain b
!! @param[out]          x the solution, i.e., Q*R'\b
!! 
subroutine dqrm_min_norm1d(qrm_mat, b, x)
  use dqrm_spmat_mod
  use qrm_mem_mod
  use qrm_error_mod
  use dqrm_analysis_mod
  use dqrm_factorization_mod
  use dqrm_solve_mod
  implicit none

  type(dqrm_spmat_type) :: qrm_mat
  real(kind(1.d0)) :: b(:), x(:)

  integer :: info
  ! error management
  integer                         :: err_act
  character(len=*), parameter     :: name='qrm_min_norm'

  call qrm_err_act_save(err_act)

  __QRM_PRNT_DBG('("Entering the least-squares driver")')

  call dqrm_check_spmat(qrm_mat)
  __QRM_CHECK_RET(name,'qrm_check_spmat',9999)
  
  if(qrm_mat%n .lt. qrm_mat%m) then
     call qrm_err_push(31, name,ied=(/qrm_mat%m,qrm_mat%n,0,0,0/))
     goto 9999
  end if

  ! analysis
  call dqrm_analyse(qrm_mat, 't')
  __QRM_CHECK_RET(name,'qrm_analyse',9999)

  ! factorization
  call dqrm_factorize(qrm_mat, 't')
  __QRM_CHECK_RET(name,'qrm_factorize',9999)

  call dqrm_solve1d(qrm_mat, 't', b, x)
  __QRM_CHECK_RET(name,'qrm_solve',9999)
  call dqrm_apply1d(qrm_mat, 'n', x)
  __QRM_CHECK_RET(name,'qrm_apply',9999)

  call qrm_err_act_restore(err_act)
  return
  
9999 continue ! error management
  call qrm_err_act_restore(err_act)
  if(err_act .eq. qrm_abort_) then
     call qrm_err_check()
  end if
  return
  
end subroutine dqrm_min_norm1d
