!! ##############################################################################################
!!
!! Copyright 2012 CNRS, INPT
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!> @file qrm_solve_mod.F90
!! This file contains a module with generic interfaces for all the solve typed routines
!!
!! $Date: 2016-01-29 22:22:30 +0100 (Fri, 29 Jan 2016) $
!! $Author: abuttari $
!! $Version: 1.1$
!! $Revision: 2075 $
!!
!! ##############################################################################################


!> @brief This module contains all the interfaces for the typed routines
!! in the solve phase
module dqrm_solve_mod

  !> @brief Generic interface for the @link ::dqrm_apply_q @endlink routine
  interface qrm_apply_q
     subroutine dqrm_apply_q(qrm_mat, b)
       use dqrm_spmat_mod
       type(dqrm_spmat_type), target :: qrm_mat
       real(kind(1.d0)), intent(inout)  :: b(:,:)
     end subroutine dqrm_apply_q
  end interface

  !> @brief Generic interface for the @link ::dqrm_apply_qt @endlink routine
  interface qrm_apply_qt
     subroutine dqrm_apply_qt(qrm_mat, b)
       use dqrm_spmat_mod
       type(dqrm_spmat_type) :: qrm_mat
       real(kind(1.d0)) :: b(:,:)
     end subroutine dqrm_apply_qt
  end interface


  !> @brief Generic interface for the
  !! @link ::dqrm_apply @endlink and
  !! @link ::dqrm_apply1d @endlink routines
  !!
  interface dqrm_apply
     module procedure dqrm_apply2dw, dqrm_apply1dw
  end interface dqrm_apply


  ! !> @brief Generic interface for the
  ! !! @link ::dqrm_apply @endlink and
  ! !! @link ::dqrm_apply1d @endlink routines
  ! !!
  interface qrm_apply
     subroutine dqrm_apply2d(qrm_mat, transp, b)
       use dqrm_spmat_mod
       type(dqrm_spmat_type), intent(in) :: qrm_mat
       real(kind(1.d0)), intent(inout)          :: b(:,:)
       character(len=*), intent(in)       :: transp
     end subroutine dqrm_apply2d
     subroutine dqrm_apply1d(qrm_mat, transp, b)
       use dqrm_spmat_mod
       type(dqrm_spmat_type), intent(in) :: qrm_mat
       real(kind(1.d0)), intent(inout)          :: b(:)
       character(len=*), intent(in)       :: transp
     end subroutine dqrm_apply1d
  end interface qrm_apply

  !> @brief Generic interface for the @link ::dqrm_solve_r @endlink routine
  interface qrm_solve_r
     subroutine dqrm_solve_r(qrm_mat, b, x)
       use dqrm_spmat_mod
       type(dqrm_spmat_type), target :: qrm_mat
       real(kind(1.d0)), intent(inout)  :: b(:,:)
       real(kind(1.d0)), intent(out)    :: x(:,:)
     end subroutine dqrm_solve_r
  end interface

  !> @brief Generic interface for the @link ::dqrm_solve_rt @endlink routine
  interface qrm_solve_rt
     subroutine dqrm_solve_rt(qrm_mat, b, x)
       use dqrm_spmat_mod
       type(dqrm_spmat_type), target :: qrm_mat
       real(kind(1.d0)), intent(inout)  :: b(:,:)
       real(kind(1.d0)), intent(out)    :: x(:,:)
     end subroutine dqrm_solve_rt
  end interface

  !> @brief Generic interface for the
  !! @link ::dqrm_solve @endlink and
  !! @link ::dqrm_solve1d @endlink routines
  !!
  interface dqrm_solve
     module procedure dqrm_solve1dw, dqrm_solve2dw
  end interface

  !> @brief Generic interface for the
  !! @link ::dqrm_solve @endlink and
  !! @link ::dqrm_solve1d @endlink routines
  !!
  interface qrm_solve
     subroutine dqrm_solve2d(qrm_mat, transp, b, x)
       use dqrm_spmat_mod
       type(dqrm_spmat_type), target :: qrm_mat
       real(kind(1.d0)), intent(inout)  :: b(:,:)
       real(kind(1.d0)), intent(out)    :: x(:,:)
       character(len=*) :: transp
     end subroutine dqrm_solve2d
     subroutine dqrm_solve1d(qrm_mat, transp, b, x)
       use dqrm_spmat_mod
       type(dqrm_spmat_type), target :: qrm_mat
       real(kind(1.d0)), intent(inout)  :: b(:)
       real(kind(1.d0)), intent(out)    :: x(:)
       character(len=*) :: transp
     end subroutine dqrm_solve1d
  end interface

  !> @brief Generic interface for the @link ::dqrm_solve_sing_front @endlink routine
  interface qrm_solve_sing_front
     subroutine dqrm_solve_sing_front(qrm_mat, b, x, trans)
       use dqrm_spmat_mod
       use dqrm_fdata_mod
       type(dqrm_spmat_type), target :: qrm_mat
       real(kind(1.d0)), intent(inout)      :: b(:,:)
       real(kind(1.d0)), intent(inout)      :: x(:,:)
       character                      :: trans
     end subroutine dqrm_solve_sing_front
  end interface

contains

  subroutine dqrm_apply2dw(qrm_mat, transp, b)
    use dqrm_spmat_mod
    type(dqrm_spmat_type), intent(in) :: qrm_mat
    real(kind(1.d0)), intent(inout)          :: b(:,:)
    character(len=*), intent(in)      :: transp
    call dqrm_apply2d(qrm_mat, transp, b)
    return
  end subroutine dqrm_apply2dw

  subroutine dqrm_apply1dw(qrm_mat, transp, b)
    use dqrm_spmat_mod
    type(dqrm_spmat_type), intent(in) :: qrm_mat
    real(kind(1.d0)), intent(inout)          :: b(:)
    character(len=*), intent(in)      :: transp
    call dqrm_apply1d(qrm_mat, transp, b)
    return
  end subroutine dqrm_apply1dw

  subroutine dqrm_solve2dw(qrm_mat, transp, b, x)
    use dqrm_spmat_mod
    type(dqrm_spmat_type), target :: qrm_mat
    real(kind(1.d0)), intent(inout)  :: b(:,:)
    real(kind(1.d0)), intent(out)    :: x(:,:)
    character(len=*) :: transp
    call  dqrm_solve2d(qrm_mat, transp, b, x)
    return
  end subroutine dqrm_solve2dw

  subroutine dqrm_solve1dw(qrm_mat, transp, b, x)
    use dqrm_spmat_mod
    type(dqrm_spmat_type), target :: qrm_mat
    real(kind(1.d0)), intent(inout)  :: b(:)
    real(kind(1.d0)), intent(out)    :: x(:)
    character(len=*) :: transp
    call dqrm_solve1d(qrm_mat, transp, b, x)
    return
  end subroutine dqrm_solve1dw


end module dqrm_solve_mod
