!! ##############################################################################################
!!
!! Copyright 2012 CNRS, INPT
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!> @file qrm_utils_mod.F90
!! This file contains a module with generic interfaces for a number of auxiliary tools
!!
!! $Date: 2016-01-29 22:22:30 +0100 (Fri, 29 Jan 2016) $
!! $Author: abuttari $
!! $Version: 1.1$
!! $Revision: 2075 $
!!
!! ##############################################################################################


!> @brief This module contains generic interfaces for a number of auxiliary tools
module dqrm_utils_mod
  use iso_c_binding

  !> @brief Generic interface for the @link ::dqrm_readmat @endlink
  !> routine
  interface qrm_readmat
     subroutine dqrm_readmat(matfile, qrm_mat, fakec)
       use dqrm_spmat_mod
       character, intent(in)                :: matfile*30
       type(dqrm_spmat_type), intent(inout) :: qrm_mat
       logical, optional                    :: fakec
     end subroutine dqrm_readmat
  end interface

  !> @brief Generic interface for the @link ::dqrm_matmul2d @endlink
  !> and @link ::dqrm_matmul1d @endlink routines
  interface dqrm_matmul
     module procedure dqrm_matmul2dw, dqrm_matmul1dw
  end interface dqrm_matmul

  !> @brief Generic interface for the @link ::dqrm_matmul2d @endlink
  !> and @link ::dqrm_matmul1d @endlink routines
  interface qrm_matmul
     subroutine dqrm_matmul2d(qrm_mat, transp, alpha, x, beta, y)
       use dqrm_spmat_mod
       type(dqrm_spmat_type), target :: qrm_mat
       real(kind(1.d0)), intent(out)        :: y(:,:)
       real(kind(1.d0)), intent(in)         :: x(:,:)
       real(kind(1.d0)), intent(in)         :: alpha, beta
       character(len=*)              :: transp
     end subroutine dqrm_matmul2d
     subroutine dqrm_matmul1d(qrm_mat, transp, alpha, x, beta, y)
       use dqrm_spmat_mod
       type(dqrm_spmat_type)         :: qrm_mat
       real(kind(1.d0)), intent(out)        :: y(:)
       real(kind(1.d0)), intent(in)         :: x(:)
       real(kind(1.d0)), intent(in)         :: alpha, beta
       character(len=*)              :: transp
     end subroutine dqrm_matmul1d
  end interface


  !> @brief Generic interface for the @link ::dqrm_vecnrm2d @endlink
  !> and @link ::dqrm_vecnrm1d @endlink routines
  interface dqrm_vecnrm
     module procedure dqrm_vecnrm2dw, dqrm_vecnrm1dw
  end interface dqrm_vecnrm

  !> @brief Generic interface for the @link ::dqrm_vecnrm2d @endlink
  !> and @link ::dqrm_vecnrm1d @endlink routines
  interface qrm_vecnrm
     subroutine dqrm_vecnrm2d(vec, n, ntype, nrm)
       real(kind(1.d0)), intent(in)  :: vec(:,:)
       real(kind(1.d0))              :: nrm(:)
       integer, intent(in)    :: n
       character              :: ntype
     end subroutine dqrm_vecnrm2d
     subroutine dqrm_vecnrm1d(vec, n, ntype, nrm)
       real(kind(1.d0)), intent(in)  :: vec(:)
       real(kind(1.d0))              :: nrm
       integer, intent(in)    :: n
       character              :: ntype
     end subroutine dqrm_vecnrm1d
  end interface

  !> @brief Generic interface for the @link ::dqrm_remap_pnt @endlink
  !> routine
  interface qrm_remap_pnt
     subroutine dqrm_remap_pnt(arr1d, pnt2d, n)
       integer :: n
       real(kind(1.d0)), target  :: arr1d(1:n)
       real(kind(1.d0)), pointer :: pnt2d(:,:)
     end subroutine dqrm_remap_pnt
  end interface


  
  interface qrm_matnrm
     subroutine dqrm_matnrm(qrm_mat, ntype, nrm)
       use dqrm_spmat_mod
       type(dqrm_spmat_type), intent(in) :: qrm_mat
       real(kind(1.d0))                         :: nrm
       character                         :: ntype
     end subroutine dqrm_matnrm
  end interface qrm_matnrm



contains

  subroutine dqrm_matmul2dw(qrm_mat, transp, alpha, x, beta, y)
    use dqrm_spmat_mod
    type(dqrm_spmat_type), target :: qrm_mat
    real(kind(1.d0)), intent(out)        :: y(:,:)
    real(kind(1.d0)), intent(in)         :: x(:,:)
    real(kind(1.d0)), intent(in)         :: alpha, beta
    character(len=*)              :: transp
    call dqrm_matmul2d(qrm_mat, transp, alpha, x, beta, y)
    return
  end subroutine dqrm_matmul2dw

  subroutine dqrm_matmul1dw(qrm_mat, transp, alpha, x, beta, y)
    use dqrm_spmat_mod
    type(dqrm_spmat_type)         :: qrm_mat
    real(kind(1.d0)), intent(out)        :: y(:)
    real(kind(1.d0)), intent(in)         :: x(:)
    real(kind(1.d0)), intent(in)         :: alpha, beta
    character(len=*)              :: transp
    call dqrm_matmul1d(qrm_mat, transp, alpha, x, beta, y)
    return
  end subroutine dqrm_matmul1dw


  subroutine dqrm_vecnrm2dw(vec, n, ntype, nrm)
    real(kind(1.d0)), intent(in)  :: vec(:,:)
    real(kind(1.d0))              :: nrm(:)
    integer, intent(in)    :: n
    character              :: ntype
    call dqrm_vecnrm2d(vec, n, ntype, nrm)
    return
  end subroutine dqrm_vecnrm2dw

  subroutine dqrm_vecnrm1dw(vec, n, ntype, nrm)
    real(kind(1.d0)), intent(in)  :: vec(:)
    real(kind(1.d0))              :: nrm
    integer, intent(in)    :: n
    character              :: ntype
    call dqrm_vecnrm1d(vec, n, ntype, nrm)
    return
  end subroutine dqrm_vecnrm1dw

end module dqrm_utils_mod
